<?php


// Contemplate Assembler (assembler.php) version 2.0.1
// Copyright 2003-2011 Arlo Leach (http://www.arlomedia.com)
// This product includes software developed by Type A Learning Agency (http://www.typea.net)
// This file may be used, modified, or redistributed under the terms
// of the open-source agreement contained in docs/license.txt


// get configuration settings

	include "../contemplate/config/constants.php";


// initialize some variables that we'll add to as we assemble the page

	$header = "";
	$output = "";
	$error_log = "";
	$embed_variable_keys = array();
	$embed_variable_values = array();
	$splits = array();
	$reserved_argument_names = array();
	$latest_mod_date = 0;	
	$latest_mod_date_formatted = "";
	

// look for runtime commands that might change our output

	$skip_processing = get_query_argument("skip_processing", "bool");
	$skip_rendering = get_query_argument("skip_rendering", "bool");
	$show_editor_links = get_query_argument("show_editor_links", "bool");
	$editor_index = get_query_argument("editor_index", "filename", "/meditate/");
	
	
// check to see if the user is logged in via basic authentication
// this is not currently working because we only see these auth variables if we're in the directory we logged into

	$login = (isset($_SERVER["PHP_AUTH_USER"])) ? $_SERVER["PHP_AUTH_USER"].":".$_SERVER["PHP_AUTH_PW"] : "" ;
					

// define our internal functions

	function get_query_string() {
		if (isset($_SERVER["QUERY_STRING"])) {
			$query_string = $_SERVER["QUERY_STRING"];
			$query_string = urldecode($query_string);
		} else {
			$query_string = "";
		}
		return $query_string;
	}

	function get_query_argument($name, $type, $default="") {
		// return the value from the $_GET array, or an empty string if the argument doesn't exist
		
		// we'll forbid users from creating arguments that might interfere with data we get through the query string
		GLOBAL $reserved_argument_names;
		$reserved_argument_names[] = $name;
		
		if (isset($_GET[$name])) {
			$value = $_GET[$name];
		} else if ($default) {
			$value = $default;
		} else {
			switch ($type) {
				case "bool": $value = false; break;
				case "name": $value = ""; break;
				case "filename": $value = ""; break;
				case "text": $value = ""; break;
			}
		}
		$value = sanitize($value, $type);

		return $value;
	}
	
	function sanitize($value, $type) {
		// remove any unnecessary characters that could be used for an exploit
		switch ($type) {
			case "bool": settype($value, "bool"); break;
			case "name": $value = preg_replace("/\W/", "", $value); break;
			case "filename": $value = preg_replace("/[^\w-\.,\/\?\&\=\% \+]/", "", $value); break;
			case "text": break;
		}
		return $value;
	}

	function get_file($file, $method="auto", $check_last_modified=0) {
		// method can be "http", "filesystem", or "auto" (auto looks at PRE_RENDER_FILES list)
		GLOBAL $output, $latest_mod_date, $latest_mod_date_formatted;
		GLOBAL $file_missing, $script_location;

		$text = "";

		if ((preg_match("/^\.\w/", $file))||(preg_match("/\/\.\w/", $file))) { // don't read files whose names start with "."
			$text = show_error("Not authorized to read the hidden file \"$file\"");
			$file_missing = 1;
		} else if ($file != sanitize($file, "filename")) { // don't read files whose names contain characters that could be used to read remote files, or to hack into the open command (Perl only)
			$text = show_error("File name \"$file\" contains characters that are not allowed");
			$file_missing = 1;
		} else {
			
			// first, figure out if we're going to read over HTTP or through the filesystem...
			if ($method == "auto") {
				$file_extension = preg_replace("/^.+\.(\w+?)$/", "$1", $file);
				$method = (in_string_array(PRE_RENDER_FILES, $file_extension)) ? "http" : "filesystem" ;
			}
			
			// ...then prepare the URL for this type of connection
			$file_URL = absolute_path($file, $method);

			// open a connection to the file
			@$myfile = fopen($file_URL, "r");
			if (!$myfile) {
				$text = show_error("Can't open the file named \"$file\"<br />".CHECK_URL);
				$file_missing = 1;
			} else {
			
				//$output .= "<p>reading $file_URL via $method</p>";  // debug

				// read the file
				while (!feof($myfile)) { $text .= fread($myfile, 4096); }
				$file_missing = 0;
				
				// close the file connection
				fclose($myfile);
				
				// for local files, catch the file's modification date (for READ_PAGE_DEFINITIONS_FROM and $template, plus content files when the template contains a env_last_modified tag)
				if ($method == "filesystem") {
					if ($check_last_modified) {
						$mtime = filemtime($file_URL);
						if ($mtime > $latest_mod_date) {
							$latest_mod_date = $mtime;
							$latest_mod_date_formatted = date("Y-m-d H:i:s", $mtime);
						}
					}
				}

				// finally, conform all line breaks to avoid cross-platform issues
				$text = conform_line_breaks($text);
	
			}

		}
		return $text;
	}
	
	function get_row($source_text, $field_name, $file_type) {
		// this is like get_field except it returns the complete HTML or XML row rather than the contents of the row
		// this is used by the content editor to update content
		GLOBAL $file_missing;
		GLOBAL $content_file;
		if (stristr($file_type, "xml")) {
			
			$source_text = make_XML_paths($source_text);
			$tag_type = "content";
			if (preg_match("/(<".$tag_type."[^>]+name=[\"\']?".preg_quote($field_name, "/")."\b[^>]*>.*?<\/$tag_type>)/is", $source_text, $matches)) {
				$row = $matches[1];
			} else {
				$row = show_error("Can't find the $tag_type tag named \"$field_name\" in $content_file<br />".CHECK_URL);
			}

		} else {
			//print "parsing row $field_name<br />"; // debug backtrack_limit

			if (!strstr($field_name, FIELD_MARKER)) {  // but what if formatting gets between the marker and the label?
				$field_name = FIELD_MARKER.$field_name;  // in that case, remove these lines and change $content_field to FIELD_MARKER.*?$content_field below
			}
			// the following should work, but the .*? before $field_name is greedy unless we precede the whole expression with .*, in which case we run over the backtrack limit
			//if (preg_match("/(<tr[^>]*>.*?<td[^>]*>.*?$field_name\W.*?\/td>.*?<td[^>]*>.*<\/td>.*?<\/tr>)/is", $source_text, $matches)) {
			//	//print "okay<br />"; // debug backtrack_limit
			//	$row = $matches[1];  // do a rough trim
			//	$row = preg_replace("/(.*?".preg_quote(FIELD_MARKER, "/").".*?)".preg_quote(FIELD_MARKER, "/").".*/is", "$1", $row, 1);  // remove any rows after this one
			//	$row = preg_replace("/(.*<\/tr>).*/is", "$1", $row, 1);  // clean up after the previous trim
			//} else {
			if (strstr($source_text, $field_name)) {
				//print "okay<br />"; // debug backtrack_limit
				$source_text = preg_replace("/$field_name\w/", FIELD_MARKER."-dont-split-here-", $source_text); // block out any field names that begin with the target field name but don't match exactly
				$halves = explode($field_name, $source_text); // split the content before and after the field name
				$rough_trim = $field_name;
				if (strstr($halves[0], FIELD_MARKER)) {
					// if there are rows before the target row, join the content after the last preceding row
					$halves[0] = explode(FIELD_MARKER, $halves[0]);
					$rough_trim = $halves[0][(count($halves[0]) - 1)].$rough_trim;
				} else {
					// otherwise just put the first half of the content back
					$rough_trim = $halves[0].$rough_trim;
				}
				if (strstr($halves[1], FIELD_MARKER)) {
					// if there are rows after the target row, join the content before the first subsequent row
					$halves[1] = explode(FIELD_MARKER, $halves[1]);
					$rough_trim = $rough_trim.$halves[1][0];
				} else {
					// otherwise just put the second half of the content back
					$rough_trim = $rough_trim.$halves[1];
				}
				// now we can do something like our original regex to clean things up
				preg_match("/.*(<tr[^>]*>.*?<td[^>]*>.*?$field_name\W.*?\/td>.*?<td[^>]*>.*<\/td>.*?<\/tr>).*/is", $rough_trim, $matches);
				$row = $matches[1];
			} else {
				//print "preg error ".preg_last_error()."<br />"; // debug backtrack_limit
				if (!$file_missing) {
					$row = show_error("Can't find the field named \"$field_name\" in $content_file<br />".CHECK_URL."<br />".CHECK_FIELD_LABEL);
				}
			}

		}
		return $row;
	}

	function update_row($row, $file_type, $new_name, $new_content) {
		// this takes a row, changes the content of the field and then returns the row code
		// this is used by the content editor to update content
		// this is easier to parse than an update_field function because here we're working with a row that is already separated from the larger content file
		GLOBAL $file_missing;
		GLOBAL $content_file;

		// conform all line breaks to avoid cross-platform issues
		// we have to do this because the content may be coming in from an external source like Meditate
		$new_content = conform_line_breaks($new_content);

		if (stristr($file_type, "xml")) {
			
			$row = make_XML_paths($row);
			$tag_type = "content";
			if (preg_match("/(<".$tag_type."[^>]+name=[\"\']?)".preg_quote($field_name, "/")."(\b[^>]*>).*(<\/$tag_type>)/is", $row, $matches)) {
				$row = $matches[1].$new_name.$matches[2].$new_content.$matches[3];
			} else {
				$row = show_error("Can't update the $tag_type tag named \"$new_name\" in $content_file<br />".CHECK_URL);
			}

		} else {		

			if (preg_match("/(<tr[^>]*>.*?<td[^>]*>.*?".preg_quote(FIELD_MARKER, "/").")\w+(\W.*?\/td>.*?<td[^>]*>).*(<\/td>.*?<\/tr>)/is", $row, $matches)) {
				$row = $matches[1].$new_name.$matches[2].$new_content.$matches[3];
			} else {
				if (!$file_missing) {
					$row = show_error("Can't update the row named \"$new_name\" in $content_file<br />".CHECK_URL."<br />".CHECK_FIELD_LABEL);
				}
			}

		}

		// change to the preferred line ending set in constants.php
		$row = str_replace("\n", LINE_ENDING_FOR_ASSEMBLED_FILES, $row);
	
		return $row;
	}
	
	function get_field_bg_color($source_text, $field_name, $file_type) {
		if (stristr($file_type, "xml")) {
			
			$source_text = make_XML_paths($source_text);
			$tag_type = "content";
			if (preg_match("/<".$tag_type."[^>]+name=[\"\']?".preg_quote($field_name, "/")."\b[^>]*bgcolor=[\"\']?([\#\w]+)[\"\']?/is", $source_text, $matches)) {
				$bg_color = $matches[1];
			} else {
				$bg_color = "";
			}

		} else {		

			if (!strstr($field_name, FIELD_MARKER)) {  // but what if formatting gets between the marker and the label?
				$field_name = FIELD_MARKER.$field_name;  // in that case, remove these lines and change $content_field to FIELD_MARKER.*?$content_field below
			}
			if (preg_match("/".$field_name."[^\w<]*<\/td>[^<]*<td[^<]*bgcolor=[\"\']?([\#\w]+)[\"\']?/is", $source_text, $matches)) {
				$bg_color = $matches[1];
			} else {
				$bg_color = "";
			}

		}
		return $bg_color;
	}
	
	function get_field($source_text, $field_name, $file_type) {
		GLOBAL $file_missing;
		GLOBAL $content_file;

		if (!$field_name) {
			// if a field name isn't specified, choose one at random
			$content_fields = get_field_names($source_text, $file_type);
			$last_field_index = count($content_fields) - 1;
			$field_name = $content_fields[rand(0, $last_field_index)];
		}

		if (stristr($file_type, "xml")) {
			
			$source_text = make_XML_paths($source_text);
			$tag_type = "content";
			if (preg_match("/<".$tag_type."[^>]+name=[\"\']?".preg_quote($field_name, "/")."\b[^>]*>(.*?)<\/$tag_type>/is", $source_text, $matches)) {
				$field_text = $matches[1];
				$field_text = str_replace("<![CDATA[", "", $field_text);
				$field_text = str_replace("]]>", "", $field_text);
			} else {
				$field_text = show_error("Can't find the $tag_type tag named \"$field_name\" in $content_file<br />".CHECK_URL);
			}

		} else {		

			if (!strstr($field_name, FIELD_MARKER)) {  // but what if formatting gets between the marker and the label?
				$field_name = FIELD_MARKER.$field_name;  // in that case, remove these lines and change $content_field to FIELD_MARKER.*?$content_field below
			}
			if (preg_match("/$field_name(\W.*\/td>.*<td[^>]*>.*<\/td>)/is", $source_text, $matches)) {
				$field_text = $matches[1];  // do a rough trim
				$field_text = preg_replace("/<\/td>/i", "", $field_text, 1);  // remove up to the previous </td>
				$field_text = preg_replace("/<td[^>]*>/is", "split-here-hey", $field_text, 1);  // mark the opening <td>
				$field_text = preg_replace("/".preg_quote(FIELD_MARKER, "/")."/i", "split-here-hey", $field_text, 1);  // mark the next field marker
				$temp = explode("split-here-hey", $field_text);  // trim at the markers
				$field_text = $temp[1];
				if (preg_match("/(.*)<\/td>/is", $field_text, $matches)) {  // remove from the trailing </td>
					$field_text = $matches[1];
				}
			} else {
				if (!$file_missing) {
					$field_text = show_error("Can't find the field named \"$field_name\" in $content_file<br />".CHECK_URL."<br />".CHECK_FIELD_LABEL);
				}
			}

		}
		return $field_text;
	}
	
	function get_form($source_text, $form_element) {
		GLOBAL $content_file;
		$value = "";
		
		// this function doesn't recognize values like this: <input name = foo> we could make it do so by adding /s* around all the equals signs, but it doesn't now
		
		// before looking for the value, let's find out if it has single or double quotes around it or not
		if (preg_match("/name=[\"\']?$form_element\W[^>]*value=[^\"\']/is", $source_text, $matches)) { $value_match = "([^ >]*)"; }
		else if (preg_match("/name=[\"\']?$form_element\W[^>]*>.*?<option[^>]+value=[^\"\']/is", $source_text, $matches)) { $value_match = "([^ >]*)"; } // the previous line covers all other input types, but the select menu is different because the type and value are in different tags (we wouldn't need this if we could write a regex that spanned the <select> and <option> tags, but that wasn't too greedy with other input types)
		else if (preg_match("/name=[\"\']?$form_element\W[^>]*value=\'/is", $source_text, $matches)) { $value_match = "'([^\']*?)'"; }
		else if (preg_match("/name=[\"\']?$form_element\W[^>]*>.*?<option[^>]+value=\'/is", $source_text, $matches)) { $value_match = "'([^\']*?)'"; } // the previous line covers all other input types, but the select menu is different because the type and value are in different tags (we wouldn't need this if we could write a regex that spanned the <select> and <option> tags, but that wasn't too greedy with other input types)
		else { $value_match = "\"([^\"]*?)\""; } // we'll let the double quote be the default, so if something goes wrong we still might get the desired result
		
		// okay, now let's look for it
		if (preg_match("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?text[\"\']?[^>]+value=$value_match"."[^>]*>/is", $source_text, $matches)) { $value = $matches[1]; }
		else if (preg_match("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?radio[\"\']?[^>]+value=$value_match"."[^>]+checked[^>]*>/is", $source_text, $matches)) { $value = $matches[1]; }
		else if (preg_match("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?checkbox[\"\']?[^>]+value=$value_match"."[^>]+checked[^>]*>/is", $source_text)) {
			// return a list (not array) of all the checked boxes
			while (preg_match("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?checkbox[\"\']?[^>]+value=$value_match"."[^>]+checked[^>]*>/is", $source_text, $matches)) {
				$value .= "$matches[1], ";
				$source_text = preg_replace("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?checkbox[\"\']?[^>]+value=$value_match"."[^>]+checked[^>]*>/is", "", $source_text, 1); // the 1 here is needed in PHP to make the replacement not global
			}
			$value = substr($value, 0, -2);
		}
		else if (preg_match("/<select[^>]+name=[\"\']?$form_element\W[^>]*>.*?<option[^>]+value=$value_match"."[^>]+selected[^>]*>.*?<\/select>/is", $source_text, $matches)) { $value = $matches[1]; }
		else if (preg_match("/<textarea[^>]+name=[\"\']?$form_element\W[^>]*>(.*?)<\/textarea>/is", $source_text, $matches)) { $value = $matches[1]; }
		else {
			if (preg_match("/name=[\"\']?$form_element\W/is", $source_text, $matches)) {
				// some HTML editors might remove the value attribute if it's empty, so maybe that's why we can't find the form element
				$value = "";
			} else {
				// otherwise, we give up
				$value = show_error("Can't find the form element named \"$form_element\" in $content_file<br />".CHECK_URL);
			}
		}
		
		return $value;
	}
	
	function get_database($database) {
		// if we're already connected to one database and the other has the same address and username, mysql_connect will automatically reuse that connection
		GLOBAL $output;
		GLOBAL ${"database_connection_".$database};
		if (!isset(${"database_connection_".$database})) {
			$address = ($database == "content") ? CONTENT_DATABASE_ADDRESS : REPORTER_DATABASE_ADDRESS ;
			$username = ($database == "content") ? CONTENT_DATABASE_USERNAME : REPORTER_DATABASE_USERNAME ;
			$password = ($database == "content") ? CONTENT_DATABASE_PASSWORD : REPORTER_DATABASE_PASSWORD ;
			$name = ($database == "content") ? CONTENT_DATABASE_NAME : REPORTER_DATABASE_NAME ;
			@${"database_connection_".$database} = mysql_connect($address, $username, $password);
			if (!${"database_connection_".$database}) {
				$output .= show_error("Could not connect to $database database");
			} else {
				@$database_selected = mysql_select_db($name);
				if (!$database_selected) {
					$output .= show_error("Could not select $database database");
					unset(${"database_connection_".$database});
				}
			}
		}
		return ${"database_connection_".$database};
	}
	
	function get_database_field($table, $column, $key_column, $key_value) {
		if (!get_database("content")) {
			$content = "";
		} else {
			if (!$table) {
				$content = show_error("Database embed tag does not contain a table argument");
			} else if (!in_string_array($table, CONTENT_DATABASE_TABLES)) {
				$content = show_error("Not authorized to read from the table \"$table\"");
			} else if (!$column) {
				$content = show_error("Database embed tag does not contain a column argument");
			} else {
				if (($key_column)&&($key_value)) {
					$query = "SELECT `$column` FROM `$table` WHERE `$key_column`='$key_value'";
				} else {
					$query = "SELECT `$column` FROM `$table` ORDER BY RAND() LIMIT 1";
				}
				$result = mysql_query($query);
				if (!$result) {
					$content = show_error("Error with SQL query");
				} else {
					$count = mysql_num_rows($result);
					if (!$count) {
						$content = show_error("Could not find database field named \"$key_value\"");
					} else {
						$row = mysql_fetch_assoc($result);
						$content = $row[$column];
						if (CONTENT_DATABASE_STRIP_SLASHES) {
							$content = stripslashes($content);
						}
					}
				}
			}
		}
		return $content;
	}

	function get_page_definitions() {
		GLOBAL $output;
		GLOBAL $page_definitions, $file_missing;
		if (!isset($page_definitions)) {
			// reading page definitions from multiple files is currently an undocumented feature
			$page_definitions = "";
			$read_page_definitions_from = READ_PAGE_DEFINITIONS_FROM;
			$read_page_definitions_from = str_replace(" ", "", $read_page_definitions_from);
			$read_page_definitions_from = explode(",", $read_page_definitions_from);
			$read_page_definitions_from_count = count($read_page_definitions_from);
			foreach ($read_page_definitions_from as $this_read_page_definitions_from) {
				$this_read_page_definitions_from = relative_path($this_read_page_definitions_from, "contemplate");
				if (stristr($this_read_page_definitions_from, "not authorized")) {
					$output .= $this_read_page_definitions_from;
				} else {
					$this_file = get_file($this_read_page_definitions_from, "filesystem", 1);
					if ($file_missing) {
						$output .= $page_definitions; // $page_definitions will contain the error message generated by get_file()
					} else {
						$page_definitions .= $this_file;
					}
				}
			}
		}
		return $page_definitions;
	}

	function get_page_definition($page_definitions, $page) {
		if (preg_match("/[^\w]".$page.":[ ]*([^\n]*)\n*/i", "\n".$page_definitions."\n", $matches)) {
			$page_definition = $matches[1];
		} else {
			$page_definition = "";
		}
		return $page_definition;
	}
	
	function get_page_name($page_definition, $source) {
		// look for a page definition that exactly matches a given dynamic URL
        $source = "\n".$source; // this makes it easier to match a page definition on the first line of the file
		$page_name = (preg_match("/[^\w]([-\w]+)\:\s+".preg_quote($page_definition, "/")."[\&\s\n]/", $source, $matches)) ? $matches[1] : "" ;
		return $page_name;
	}
	
	function get_page_names_containing($args, $source) {
		// look for page definitions that contain a given set of arguments
		$pages = Array();
		$source = "\n".$source; // this makes it easier to match a page definition on the first line of the file
		while (preg_match("/[^\w]([-\w]+)\:\s+[^\n]+=".preg_quote($args, "/")."[^\n]*[\&\s\n]/", $source, $matches)) {
			array_push($pages, $matches[1]);
			$source = preg_replace("/([^\w])[-\w]+\:\s+[^\n]+=".preg_quote($args, "/")."[^\n]*[\&\s\n]/", "$1", $source, 1);
		}
		return $pages;
	}
	
	function get_page_names($source, $grouped=0) {
		// return a list of all the page names in a page definitions file
		$page_names = Array();
		$comment_pattern = "/#[ ]*\w+\\n/";
		if (($grouped)&&(preg_match_all($comment_pattern, $source, $matches, PREG_PATTERN_ORDER))) {
			// if we have comments in the file, return the page names as an array of arrays with the comments as the keys of the outer array
			// a comment is defined as a line starting with #, followed by optional spaces, followed by only alphanumeric characters
			$groups = array();
			$grouped_source = preg_split($comment_pattern, $source);
			array_shift($grouped_source); // lose anything before the first comment
			$group_names = $matches[0];
			$groups_count = count($group_names);
			for ($i=0; $i<$groups_count; $i++) {
				$this_group_name = $group_names[$i];
				$this_group_name_cleaned = preg_replace("/\W/", "", $this_group_name);
				$groups[$this_group_name_cleaned] = $grouped_source[$i];
			}
		} else {
			$groups = array("pages"=>$source);
		}
		foreach ($groups as $group_name=>$group_source) {
			$page_names[$group_name] = array();
			while (preg_match("/([-\w]+)\:\s+[\w]+/", $group_source, $matches)) {
				$this_page_name = $matches[1];
				array_push($page_names[$group_name], $this_page_name);
				$group_source = str_replace($this_page_name.":", "", $group_source);
			}
		}
		/*
		if (count($page_names) == 1) {
			// if there's only one group, return a single array instead of an array of arrays
			// actually, this requires similar logic on the receiving end, so it's probably best just to return everything consistently
			foreach ($page_names as $group_name=>$group_source) {
				$page_names = $group_source;
			}
		}
		*/
		return $page_names;
	}
	
	function update_page_definition($name, $source, $args) {
		$source = preg_replace("/($name: )[^\\n]*(\\s?)/i", "$1$args$2", $source);
		return $source;
	}
	
	function copy_page_definition($name_original, $name_new, $source) {
		$source = preg_replace("/($name_original: )([^\\n]*\\s?)/i", "$1$2$name_new: $2", $source);
		return $source;
	}
	
	function rename_page_definition($name_original, $name_new, $source) {
		$source = preg_replace("/$name_original: ([^\\n]*\\s?)/i", "$name_new: $1", $source);
		return $source;
	}
	
	function delete_page_definition($name, $source) {
		$source = preg_replace("/$name: [^\\n]*\\s?/i", "", $source);
		return $source;
	}
	
	function get_field_names($source, $file_type) {
		// return a list of all the content fields in a content file -- either HTML or XML depending on the second argument
		$field_names = Array();

		if (stristr($file_type, "xml")) {

			$source = make_XML_paths($source);
			$tag_type = "content";
			while (preg_match("/<".$tag_type."[^>]+name=[\"\']?([-\w\/]+)[\"\']?[^>]*>/is", $source, $matches)) {
				$this_field_name = $matches[1];
				array_push($field_names, $this_field_name);
				$source = preg_replace("/<".$tag_type."[^>]+name=[\"\']?[-\w\/]+[\"\']?[^>]*>/is", "", $source, 1);
			}

		} else {

			while (preg_match("/".preg_quote(FIELD_MARKER, "/")."([-\w]+)/", $source, $matches)) {
				$this_field_name = $matches[1];
				array_push($field_names, $this_field_name);
				$source = preg_replace("/".preg_quote(FIELD_MARKER, "/")."/", "", $source, 1); // can't be str_replace because it must only replace once
			}

		}
		return $field_names;
	}
	
	function get_form_names($source) {
		// return a list of all the form elements in a content field
		$form_names = Array();
		while (preg_match("/<input[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is", $source, $matches)) {
			$this_form_name = $matches[1];
			array_push($form_names, $this_form_name);
			$source = preg_replace("/".preg_quote($matches[0], "/")."/", "", $source, 1); // we can't use str_replace here because we must only replace once
		}
		while (preg_match("/<select[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is", $source, $matches)) {
			$this_form_name = $matches[1];
			array_push($form_names, $this_form_name);
			$source = preg_replace("/".preg_quote($matches[0], "/")."/", "", $source, 1); // we can't use str_replace here because we must only replace once
		}
		while (preg_match("/<textarea[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is", $source, $matches)) {
			$this_form_name = $matches[1];
			array_push($form_names, $this_form_name);
			$source = preg_replace("/".preg_quote($matches[0], "/")."/", "", $source, 1); // we can't use str_replace here because we must only replace once
		}
		return $form_names;
	}
	
	function make_XML_paths($source) {
		// rewrite nested XML content as flat content with paths for tag names; assumes no more than one tag per line
		$source = explode("\n", $source);
		$path = Array();
		for ($n=0; $n < count($source); $n++) {
			$this_line = $source[$n];
			if (preg_match("/<group[^>]+name=[\"\']?([-\w]+)[\"\']?[^>]*>/is", $this_line, $matches)) {
				$group_name = $matches[1];
				array_push($path, $group_name); // add the new group name to the path array
			} else if (strstr($this_line, "</group>")) {
				$group_name = array_pop($path); // remove the latest group name from the path array
			} else if (preg_match("/<content[^>]+name=[\"\']?([-\w]+)[\"\']?[^>]*>/is", $this_line, $matches)) {
				$field_name = $matches[1];
				if (count($path) > 0) {
					$current_path = implode("/", $path);
					$field_name = $current_path."/".$field_name;
					$this_line = preg_replace("/(<content[^>]+name=[\"\']?)[-\w]+([\"\']?[^>]*>)/is", "$1".$field_name."$2", $this_line, 1);
					$source[$n] = $this_line;
				}
			}
		}
		$source = implode("\n", $source);
		return $source;
	}
	
	function get_directory_listing($directory) {
		// return a list of all the items in the specified directory
		if (file_exists(absolute_path($directory, "filesystem"))) {
			if ($dir = opendir(absolute_path($directory, "filesystem"))) {
				$files = Array();
				while ($file = readdir($dir)) {
					if (substr($file, 0, 1) != ".") { array_push($files, $file); } // that gets files that don't start with a dot
				}  
				closedir($dir);
			}
		}
		return $files;
	}
	
	function get_recursive_listing($start_folder, $recursive=0, $filter="") {
		// return a list of all the files in a given folder, including subfolders
		$start_folder = preg_replace("/[\\/\\\]$/", "", $start_folder, 1); // remove a trailing slash
		$file_list = Array();
		$folder_list = Array($start_folder);
		while ($folder_list) {
			$this_folder_path = $folder_list[0];
			$this_folder_items = get_directory_listing($this_folder_path);
			for ($n=0; $n<count($this_folder_items); $n++) {
				$this_item_name = $this_folder_items[$n];
				$this_item_path = $this_folder_path."/".$this_item_name;
 				// if the item name contains a dot, we assume it's a file, if not, we assume it's a directory
 				if (!strstr($this_item_name, ".")) {
 					// we've found a folder; add it to the list of folders to scan
 					if ($recursive) {
						array_push($folder_list, $this_item_path);
					}
				} else if ((!$filter)||(strstr($this_item_name, $filter))) {
					// we've found a file; add it to the list of files to return
					$this_item_path = str_replace($start_folder."/", "", $this_item_path); // make results relative to the start folder, not to the site
					array_push($file_list, $this_item_path);
				}
			}
			array_shift($folder_list); // remove the folder we just scanned from the list
		}
		return $file_list;
	}
	
	function process_tag_nav($this_type) {
		GLOBAL $page_definitions, $page_definition, $page;
		
		// if we came here from the expanded URL and not the page shortcut, get the page definitions and page name now
		if (!$page) {
			$page_definitions = get_page_definitions();
			$page = get_page_name($page_definition, $page_definitions);
		}

		$nav_page = $page;
		
		$nav_pages = Array();
		$nav_page_position = 0;
		$count_pages = 0;
		
		// find the block of pages (separated by double line breaks) that this page belongs to
		// first, split the page definitions file up into blocks
		$pages_blocks = "\n\n$page_definitions\n\n";
		$pages_blocks = str_replace("\n\n", "\n\n\n", $pages_blocks);
		if (preg_match("/.*\n\n(.*?\W".$nav_page.":[ ]*.*?)\n\n.*/s", $pages_blocks, $matches)) {
			// then find the block that contains this page
			$nav_block = $matches[1];
			$nav_block = preg_replace("/\:[ ]*[^\n]*/", "", $nav_block, -1);
			$nav_block = trim($nav_block);
			$nav_pages = explode("\n", $nav_block);
			$count_pages = count($nav_pages);
		}
		for ($j=0; $j<$count_pages; $j++) {
			// finally, parse through it to find this page's position in the block
			if (strtolower($nav_pages[$j]) == strtolower($nav_page)) {
				$nav_page_position = $j+1;
				$j = $count_pages; // stop the loop
			}
		}
					
		// don't bother with the next part, if the current page doesn't exist
		if (!$nav_page_position) {
			$content = show_error("Can't find the page named \"$nav_page\" in ".READ_PAGE_DEFINITIONS_FROM);
		} else {
	
			if (strstr($this_type, "nav_current")) {
				$content = round($nav_page_position); // return the position of the current page
			} else if (strstr($this_type, "nav_total")) {
				$content = $count_pages; // return the number of pages in this block
			} else if (strstr($this_type, "nav_first")) {
				$content = $nav_pages[0]; // return the name of the first page
			} else if (strstr($this_type, "nav_last")) {
				$content = $nav_pages[$count_pages - 1]; // return the name of the last page
			} else if (strstr($this_type, "nav_next")) {
				$result_page_position = $nav_page_position + 1;
				if ($result_page_position > $count_pages) { $result_page_position = 1; }
				$content = $nav_pages[$result_page_position - 1]; // return the name of the requested page
			} else if (strstr($this_type, "nav_previous")) {
				$result_page_position = $nav_page_position - 1;
				if ($result_page_position < 1) { $result_page_position = $count_pages; }
				$content = $nav_pages[$result_page_position - 1]; // return the name of the requested page
			}
				
		}
		
		return $content;
	}
	
	function process_tag_env($this_type) {
		GLOBAL $latest_mod_date_formatted;
		preg_match("/env_(.+)/i", $this_type, $matches);
		$variable = $matches[1];
		if (stristr($variable, "last_mod")) {
			$content = $latest_mod_date_formatted;
		} else {
			$content = (isset($_SERVER[strtoupper($matches[1])])) ? $_SERVER[strtoupper($matches[1])] : "" ;
			$content = stripslashes($content); // these get escaped in PHP for some reason
		}
		return $content;
	}

	function show_error($message) {
		GLOBAL $output, $error_log;
		
		// prepare the shown message by removing quote marks
		$message_show = $message;
		$message_show = preg_replace("/\"/", "&quot;", $message_show, -1);
		$message_show = preg_replace("/'/", "&#039;", $message_show, -1);
		$message_show = "<div class=\"assembler_error\" style=\"".ERROR_BOX_STYLE."\"><strong>Assembler error</strong><br />$message_show</div>";
		
		// prepare the hidden message by removing <br /> tags
		$message_hide = $message_show;
		$message_hide = str_replace("<br />", ";", $message_hide);
		$message_hide = "<!-- Assembler error: $message_hide -->";
		
		// prepare the logged message by removing all but the first line
		$message_log = $message;
		if (strpos($message_log, "<br />")) {
			$message_log = substr($message_log, 0, strpos($message_log, "<br />"));
		}
		$message_log = strip_tags($message_log); // also remove any HTML formatting from it
		
		// add the log message to the ongoing error log
		if ($error_log) { $error_log .= ","; } // and comma-delimit multiple errors on a page
		$error_log .= $message_log;
		
		// add the shown and hidden messages to the output
		if (SHOW_ERRORS) {
			// only display one error at a time, since subsequent errors are usually a result of the first one
			if (!strstr($output, "Assembler error")) {
				return $message_show;
			} else {
				return "";
			}
		} else {
			return $message_hide;
		}
		
	}
	
	function relative_path($path, $subfolder="") {
		// formats an initial reference to a file and authorizes it for reading
		// these two functions could probably be combined because we always call one and then the other
		GLOBAL $output;
		GLOBAL $login;
		
		$error = "";

		// don't allow files from other servers
		$path = str_replace("://", "", $path);
		
		// if the path (before we add the subfolder) contains any folder names, we'll validate it against the authorized locations
		$validate_location = (strstr($path, "/"));

		// if the path doesn't include a subfolder, add the subfolder we passed in
		if ((substr($path,0,3) != "../")&&(substr($path,0,2) != "./")) {
			if ($subfolder) { $subfolder .= "/"; }
			$path = "../".$subfolder.$path;
		}

		// if required, check the path name against the authorized locations listed in the settings file
		if ($validate_location) {
			$authorized = 0;
			if (READ_FILES_FROM) {
				$authorized = in_string_array($path, READ_FILES_FROM, 1);
			}
			//$output .= "<p>authorized $path: $authorized</p>";  // debug
			if (!$authorized) {
				if (preg_match("/(.*\/)[^\/]*/i", $path, $matches)) { $path_trimmed = $matches[1]; }
				$error = show_error("Not authorized to read from the path \"$path_trimmed\"<br />".CHECK_PATHS);
			}
		}

		// see if we require authentication to view this file and show an error if necessary
		if ((!$error)&&(in_string_array($path, REQUIRE_AUTHENTICATION_TO_READ_FILES_FROM, 1))) {
			if (!$login) {
				if (preg_match("/(.*\/)[^\/]*/i", $path, $matches)) { $path_trimmed = $matches[1]; }
				$error = show_error("As of Contemplate version 2.0, the REQUIRE_AUTHENTICATION_TO_READ_FILES_FROM option is known to be broken; please contact us if you would like to use this functionality");
				//$error = show_error("Not authorized to read from the path \"$path_trimmed\" without logging in");
				//$header .= $_SERVER["SERVER_PROTOCOL"]." 403 Not Authorized\n"; // this sends the status for PHP/mod_apache
				//$header .= "Status: 403 Not Authorized\n"; // this sends the status for PHP/CGI
				////$_SERVER["REDIRECT_STATUS"] = "403"; // this tells the reporter to log this as an error page
				// could we show the login box here instead of an error page? http://php.net/manual/en/features.http-auth.php
			}
		}

		if ($error) {
			return $error;
		} else {
			return $path;
		}
	}
	
	function absolute_path($path, $method) {
		// formats a complete reference to a file for reading via the filesystem or http
		// these two functions could probably be combined because we always call one and then the other
		GLOBAL $output;
		GLOBAL $script_location, $request, $URL_rewriting_flag;

		if ($method == "filesystem") {
		
			if (!$script_location) {
				$output .= show_error("Couldn't determine the filesystem path to this script. Please ensure that either the SCRIPT_FILENAME or PATH_TRANSLATED variable is available on your server.");
			} else {
				$script_location_copy = $script_location;
				$file_copy = $path;
				while (strstr($file_copy, "../")) {
					// this avoids results like contemplate/../templates/default.html
					$script_location_copy = preg_replace("/[^\/]+\/$/", "", $script_location_copy, 1); // remove the last directory name from the script location
					$file_copy = preg_replace("/\.\.\//", "", $file_copy, 1); // remove the first updelim from the filename (in PHP, we need to use preg_replace so we only do one replacement)
				}
				$path = $script_location_copy.$file_copy;
			}
			
		} else {
		
			if (!$request) {
				$output .= show_error("Couldn't determine the web path to this script. Please ensure that either the SCRIPT_NAME or REQUEST_URI variable is available on your server.");
			} else {
				$thisserver = (isset($_SERVER["HTTP_HOST"])) ? $_SERVER["HTTP_HOST"] : "" ;
				$thispath = "";
				if ($URL_rewriting_flag) {
					if (preg_match("/\/([^=]*\/)$URL_rewriting_flag\//i", $request, $matches)) { $thispath = $matches[1]; }
				} else {
					if (preg_match("/\/([^=]*\/)contemplate\//i", $request, $matches)) { $thispath = $matches[1]; }
				}
				
				// pass along the authentication information if present
				$login = (isset($_SERVER["PHP_AUTH_USER"])) ? $_SERVER["PHP_AUTH_USER"].":".$_SERVER["PHP_AUTH_PW"]."@" : "" ; 
				
				$path = preg_replace("/\.\.\//", "http://$login$thisserver/$thispath", $path, 1); // replace the first updelim with the scheme and path (in PHP, we need to use preg_replace so we only do one replacement)
				$path = str_replace("../", "", $path); // remove additional updelims
			}
			
		}
		
		return $path;
	}

	function make_preview($before_str, $str, $after_str) { // return a little bit before and a little bit after the selected text (used for search results)
		$ellipses_before = (strlen($before_str) > SEARCH_RESULTS_PREVIEW_LENGTH) ? "..." : "" ;
		$ellipses_after = (strlen($after_str) > SEARCH_RESULTS_PREVIEW_LENGTH) ? "..." : "" ;
		$before_str = substr($before_str, -(round(SEARCH_RESULTS_PREVIEW_LENGTH / 2)));
		$after_str = substr($after_str, 0, round(SEARCH_RESULTS_PREVIEW_LENGTH / 2));
		$before_str = preg_replace("/^\s+/", "", $before_str);
		$after_str = preg_replace("/\s+$/", "", $after_str);
		$preview = $ellipses_before.$before_str."<strong>$str</strong>".$after_str.$ellipses_after;
		return $preview;
	}

	function conform_line_breaks($text) {
		if (strpos($text, "\r") !== false) {
			$text = str_replace("\r\n", "\n", $text); // change DOS to Unix
			$text = str_replace("\r", "\n", $text); // change Mac to Unix
		}
		return $text;
	}

	function stri_replace($pattern, $replacement, $subject) { // do a case-insensitive replace; this is a substitute for a PHP function that doesn't exist!
		return preg_replace("/".preg_quote($pattern, "/")."/is", $replacement, $subject, -1);
	}
	
	function in_string_array($value, $string_array, $recursive=0) {
		// see if a value is contained within a comma-delimited list
		// this saves us having to change several of our constants into local array variables, since constants can't be arrays
		if ($string_array) {
			if (!$recursive) {
				if (preg_match("/[, ]{$value}[, ]/", ",$string_array,")) { return true; }
			} else {
				$array = explode(",", str_replace(" ", ",", $string_array));
				foreach ($array as $item) {
					if (strpos($value, $item) === 0) { return true; }
				}
			}
		}
		return false;
	}

	function random_string($strlen) {
		$chars = "0123456789bcdfghjklmnpqrstvwxyzBCDFGHJKLMNPQRSTVWXYZ";
		$str = "";
		for ($n=0; $n<$strlen; $n++) {
			$str .= substr($chars, rand(0, (strlen($chars) - 1)), 1);
		}
		return $str;
	}
	

// define some functions that would be useful to users

	function merge_values($_content, $values) {
		// if $content contains any PHP variables ($var), this function will replace them with the values included in the associative array $data
		// tip: if you don't have all your desired values collected into an associative array, you can pass $GLOBALS as the second argument
		extract($values);
		eval('$_content = stripSlashes("'.addSlashes($_content).'");');
		return $_content;
	}


// set up some primary variables

	// try to figure out the path to this script as it appears in the browser, before any URL rewriting
	// in IIS, this probably won't work, because REQUEST_URI will either not exist, or will show the path after rewriting, since rewriting is handled externally from IIS
	$request = "";
	if (isset($_SERVER["REQUEST_URI"])) {
		$request = $_SERVER["REQUEST_URI"];
		if (strpos($request, "?") !== false) {
			$request = substr($request, 0, strpos($request, "?"));
		}
	} else {
		// IIS doesn't always provide the REQUEST_URI variable
		if (isset($_SERVER["SCRIPT_NAME"])) {
			$request = $_SERVER["SCRIPT_NAME"];
		}
	}
	
	// if we know the browser path, figure out the part that triggers the rewrite; assume this is just one virtual directory name
	$URL_rewriting_flag = "";
	if (strpos($request, "contemplate/assembler") === false) {
		if (preg_match("/([^\/]*)\/[^\/]+$/i", $request, $matches)) {
			$URL_rewriting_flag = $matches[1];
		}
	}

	// figure out the full local path of the contemplate directory so we can build an absolute path that works on Windows computers
	// on Unix, this will include any symlinks used to construct the URL, so we can run the script from a symlink to the real contemplate folder, 
	// and paths will be relative to the apparent, rather than the actual, location of assembler.asp
	// on Windows, this will not include virtual directories, but constructing a path from a virtual directory wouldn't work anyway
	// because those aren't available at the file system level
	// we could save some processing time by hard-coding this value into the constants.php file, but this is more convenient for people who don't know their directory paths or who want to move their site from one directory to another
	$script_location = "";
	if (isset($_SERVER["SCRIPT_FILENAME"])) {
		$script_location = $_SERVER["SCRIPT_FILENAME"];
	} else {
		// Windows doesn't provide the SCRIPT_FILENAME variable
		if (isset($_SERVER["PATH_TRANSLATED"])) {
			$script_location = $_SERVER["PATH_TRANSLATED"];
		}
	}
	$script_location = str_replace("\\", "/", $script_location);
	$script_location = str_replace("//", "/", $script_location); // not sure why PHP returns double slashes on Windows
	if (($URL_rewriting_flag)&&(preg_match("/(.*\/)$URL_rewriting_flag\//i", $script_location, $matches))) {
		$script_location = $matches[1]."contemplate/";
	} else {
		$script_location = preg_replace("/(.*\/)[^\/]+/", "$1", $script_location, 1);
	}


// if we're running as an include file for the sake of our functions only, stop processing now

	if ($skip_processing) {
		return;
	}


// write an HTTP header to prepare for text output

	$header .= "Content-type: text/html\n";
	if (PREVENT_PAGE_CACHING) {
  		// trick the browser into getting a new page every time
  		$header .= "Pragma: no-cache\n";
		$header .= "Cache-control: no-cache\n";
		$header .= "Expires: Mon, 07 May 1973 00:00:00 GMT\n";
	}
	if ((USE_COOKIES_TO_TRACK_VISITORS)&&(!isset($_COOKIE["Contemplate"]))) {
		// if we want to track visitors and this visitor doesn't have a Contemplate cookie yet, generate a visitor ID and save it to the cookie (this assumes that if they have the cookie, it contains a visitor ID)
		$SQL_timestamp_stripped = preg_replace("/\D/", "", date("YmdHis"));
		$visitor_id = random_string(16)."-".$SQL_timestamp_stripped;
		$header .= "Set-Cookie: Contemplate=visitor_ID%3D$visitor_id; Expires=".date("D, d M Y 00:00:00 \G\M\T", time("") + (60*60*24*365*10))."; Path=/; Version=1";
	}


// parse the URL to figure out the arguments for this page

	$page = get_query_argument("page", "name");
	$requested_page = get_query_argument("requested_page", "name"); // this is needed for the SHOW_EXPANDED_URLS option
	if ($page) {
		// look in the page definitions...
		$page_definitions = get_page_definitions();
		$page_definition = get_page_definition($page_definitions, $page);
		if ($page_definition) { // CHECK FOR THE PRESENCE OF THE TEMPLATE VALUE ELSEWHERE (??)
			// let any real-time args tag along for the ride
			$more_args = get_query_string();
			$more_args = preg_replace("/&?page=[^&]+/", "", $more_args, 1);
			if ($more_args) { $page_definition .= "&".$more_args; }
			$page_definition = preg_replace("/&+/", "&", $page_definition, -1);
			if (SHOW_EXPANDED_URLS) {
				// redirect to the page with the full URL visible
				$page_definition = stri_replace("&show_expanded_URLs=1", "", $page_definition, 1);
				$page_definition .= "&requested_page=".$page;
				header("Location: ../contemplate/assembler.php?".$page_definition);
				exit;
			}
		} else {
			$read_page_definitions_from_plural = (strpos(READ_PAGE_DEFINITIONS_FROM, ",") !== false) ? "s" : "" ;
			$error_message = show_error("Can't find a definition for \"$page\" in the file$read_page_definitions_from_plural ".READ_PAGE_DEFINITIONS_FROM."."); // running this function here not only formats the error box if we need it, but adds the error text to the error log
			$header .= $_SERVER["SERVER_PROTOCOL"]." 404 Not Found\n"; // this sends the status for PHP/mod_apache
			$header .= "Status: 404 Not Found\n"; // this sends the status for PHP/CGI
			// and this actually shows the error page (setting the status doesn't trigger Apache to show it)
			if (!ERROR_PAGE_404) { // nothing specified
				$output .= $error_message;
			} else if (preg_replace("/[.\/]/", "", ERROR_PAGE_404) == ERROR_PAGE_404) { // looks like a page definition
				$page_definition = get_page_definition($page_definitions, ERROR_PAGE_404);
				if ($page_definition) {
					$_SERVER["REDIRECT_STATUS"] = "404"; // this tells the reporter to log this as an error page
				} else {
					$output .= $error_message;
				}
			} else { // looks like another page
				$header .= "Location: ".ERROR_PAGE_404."\n";
			}
		}
	} else {
		// ...or just use the current URL
		$page_definition = get_query_string();
		if ($requested_page) {
			$page = $requested_page;
			$page_definition = stri_replace("&requested_page=$page", "", $page_definition, 1);
		}
	}
	
	// don't proceed from here unless we have a page definition
	if ($page_definition) {
	
		// parse the page definition to get the individual arguments
		$page_definition = sanitize($page_definition, "filename");
		$page_definition_array = explode("&", $page_definition);
		$args = array();
		foreach ($page_definition_array as $page_definition_item) {
			if (strpos($page_definition_item, "=") !== false) {
				// get the key and value for each element of the page definition
				list($key, $value) = explode("=", $page_definition_item);

				// make argument names case-insensitive
				$key = strtolower($key);
	
				if ($key == "template") {
					// only this argument doesn't get turned into an array; but we do make it available as a local variable
					${$key} = $value;
				} else {
					if (strstr($value, ",")) {
						// turn the elements of the argument value into an array so we can more easily refer to its individual parts
						$args[$key] = explode(",", $value);
					} else {
						// if the value only has one element, turn it into an array anyway for consistency
						$args[$key] = Array($value);
					}
				}
			}
		}
		
	
// grab the text of the template file
	
		if (isset($template)) {
			$template_text = relative_path($template, "templates");
			if (!stristr($template_text, "not authorized")) { // if there's an error, leave the error in the template_text variable to display it
				$template_text = get_file($template_text, "auto", 1);
			}
		} else {
			$template = "";
			$template_text = "<!--#embed args -->"; // allow display of one piece of content directly, with no template
		}
		$output .= $template_text;
		
		
// make the core processing recursive (e.g. process embed tags in embedded content)

		$i = 1;
		while (preg_match("/<!--\#\s*embed.*-->/is", $output)||(preg_match("/<!--\#\s*if.*-->/is", $output))) {
		

// process any conditional tags to hide pieces of the template
	
			// first, find and number all if/else/end if sets
			$levels = Array();
			$greatest = 0;
			while (preg_match("/<!--\#\s*end[ ]*if\D/is", $output)) {
				// which is next, an if, else, or end if?
				if (preg_match("/<!--\#\s*if\D(.*)/is", $output, $matches)) { $length_after_if = strlen($matches[1]); } else { $length_after_if = -1; }
				if (preg_match("/<!--\#\s*else\D(.*)/is", $output, $matches)) { $length_after_else = strlen($matches[1]); } else { $length_after_else = -1; }
				if (preg_match("/<!--\#\s*end[ ]*if\D(.*)/is", $output, $matches)) { $length_after_end_if = strlen($matches[1]); } else { $length_after_end_if = -1; }
				if (($length_after_if > $length_after_end_if)&&($length_after_if > $length_after_else)) {
					$greatest++;
					array_push($levels, $greatest);
					$output = preg_replace("/<!--\#\s*if(\D)/i", "<!--#if$greatest$1", $output, 1);
					//$output .= "<br />adding if $greatest";  // debug
				} else if (($length_after_else > $length_after_if)&&($length_after_else > $length_after_end_if)) {
					$next = $levels[count($levels) - 1];
					$output = preg_replace("/<!--\#\s*else(\D)/i", "<!--#else$next$1", $output, 1);
					//$output .= "<br />adding else $next";  // debug
				} else {
					$next = array_pop($levels);
					$output = preg_replace("/<!--\#\s*end[ ]*if(\D)/i", "<!--#end if$next$1", $output, 1);
					//$output .= "<br />adding end if $next";  // debug
				}
			}
			
			// then evaluate and process each of them
			while (preg_match("/<!--\#\s*if(\d+)[ ]*\((.*?[^ ])\)[ ]*-->/is", $output, $matches)) {  // ignore extra spaces around "if" and expression
				$j = $matches[1];
				$this_expression = $matches[2];
				$native_expression = $this_expression;
				
				// the variable symbol $ is optional in the template; add it here if it's missing
				if (!preg_match('/\$[_\w]+\[/', $native_expression)) { $native_expression = preg_replace("/([_\w]+\[)/", "$"."$1", $native_expression, -1); }
				
				// these variables don't exist internally as arrays; this fakes it so they're still available in conditional tags
				if ((SPLIT_PAGE_NAMES_ON)&&(preg_match('/\$page\\[\\d+\\]/i', $native_expression))) {
					$page_array = explode(SPLIT_PAGE_NAMES_ON, $page);
					$native_expression = preg_replace('/\$page\[(\d+)\]/i', "\$page_array[$1]", $native_expression);
				} else {
					$native_expression = preg_replace('/\$page\[\d*\]/i', "\$page", $native_expression);
				}
				$native_expression = preg_replace('/\$template\[\d*\]/i', "\$template", $native_expression);
				$native_expression = preg_replace('/\$page_definition\[\d*\]/i', "\$page_definition", $native_expression);

				// if an array (other than the special variables above) doesn't have an index, return every element of the array as a string
				$native_expression = preg_replace('/\\$([_\\w]+)\\[\\]/i', "implode(\",\", \$$1)", $native_expression);
				
				// these keywords don't exist as arrays at all; this inserts their values as literal strings so they're available, too
				while (preg_match('/\$*(env_\w+)/i', $native_expression, $matches)) {
					$native_expression = preg_replace('/\$*env_\w+/i', "\"".process_tag_env($matches[1])."\"", $native_expression, 1);
				}
				while (preg_match('/\$*(nav_\w+)/i', $native_expression, $matches)) {
					$native_expression = preg_replace('/\$*nav_\w+/i', "\"".process_tag_nav($matches[1])."\"", $native_expression, 1);
				}

				// if a variable in the expression doesn't exist, look for it in our arguments array
				// if it still doesn't exist, replace it with an empty string to avoid "unset" errors when we evaluate the expression
				while (preg_match('/\$([_\w]+)\[([^\]])\]/i', $native_expression, $matches)) {
					if (isset(${$matches[1]})&&(count(${$matches[1]}) > $matches[2])) {
						$native_expression = str_replace("\$".$matches[1]."[".$matches[2]."]", FIELD_MARKER.$matches[1]."[".$matches[2]."]", $native_expression);
					} else if (isset($args[$matches[1]])&&(count($args[$matches[1]]) > $matches[2])) {
						$native_expression = str_replace("\$".$matches[1]."[".$matches[2]."]", FIELD_MARKER."args['".$matches[1]."'][".$matches[2]."]", $native_expression);
					} else {
						$native_expression = str_replace("\$".$matches[1]."[".$matches[2]."]", "\"\"", $native_expression);
					}
				}				
				$native_expression = str_replace(FIELD_MARKER, "$", $native_expression); // not sure why we had to do that...
				
				//$output .= $native_expression."<br />"; // debug
				eval('$this_evaluates = ('.$native_expression.');');

				if (preg_match("/<!--\#\s*else".$j."\D/i", $output)) {
					if ($this_evaluates) {
						$output = preg_replace("/<!--\#\s*if".$j."[ ]*\((.*?[^ ])\)[ ]*-->/is", "", $output, 1); // remove the if tag...
						$output = preg_replace("/<!--\#\s*else".$j."\D.*end[ ]*if".$j."[ ]*-->/is", "", $output, 1); // ...and from the else tag to the end if tag
					} else {
						$output = preg_replace("/<!--\#\s*if".$j."\D.*else".$j."[ ]*-->/is", "", $output, 1); // remove from the if tag to the else tag...
						$output = preg_replace("/<!--\#\s*end[ ]*if".$j."[ ]*-->/is", "", $output, 1); // ...and the end if tag
					}
				} else {
					if ($this_evaluates) {
						$output = preg_replace("/<!--\#\s*if".$j."[ ]*\((.*?[^ ])\)[ ]*-->/is", "", $output, 1); // remove the if tag
						$output = preg_replace("/<!--\#\s*end[ ]*if".$j."[ ]*-->/is", "", $output, 1); // remove the end if tag
					} else {
						$output = preg_replace("/<!--\#\s*if".$j."\D.*end[ ]*if".$j."[ ]*-->/is", "", $output, 1); // remove from the if to the end if tag
					}
				}
			}
		

// find all the embed tags in the template and replace with temporary placeholders
		
			// slower: while (preg_match("/(.*?)<!--\#\s*embed\s*([^\n]*?[^ ])\s*-->(.*)/is", $output, $matches)) {  // remove extra spaces before embed type, around "embed", and after attributes
			while (preg_match("/(<!--\#\s*embed\s*)([^\n]*?[^ ])(\s*-->)/is", $output, $matches)) {  // ignore extra spaces before embed type, around "embed", and after attributes
			
				// slower: $before_embed = $matches[1]; $this_embed = $matches[2]; $after_embed = $matches[3];
				$this_embed_complete = $matches[1].$matches[2].$matches[3];
				$this_embed = $matches[2];
				$this_embed = strtolower($this_embed); // make embed tags case-insensitive
				$this_embed_copy = $this_embed;
				$this_type = "";
				$this_embed_variable = "";
				$this_embed_strip = "";
				$this_embed_replace = Array();
				$this_embed_passthru_index = "";
				
				if (preg_match("/\s+variable/", $this_embed)) {  // look for a variable attribute, which tells us to save the embed content as a variable rather than add it to the output
					$this_embed_variable = 1;
					$this_embed = preg_replace("/\s+variable/", "", $this_embed, 1);
				}
				
				if (preg_match("/\s+strip/", $this_embed)) {  // look for a strip attribute, which tells us to remove all markup from the embed content
					$this_embed_strip = 1;
					$this_embed = preg_replace("/\s+strip/", "", $this_embed, 1);
				}
				
				if (preg_match("/\s+replace/", $this_embed)) {  // look for a replace attribute, which tells us to do some search-and-replaces on the embed content
					$this_replace_array = Array();
					$this_embed = str_replace("\\\\", "-escaped-backslash-", $this_embed); // temporarily replace escaped backslashes so we can access the escaped slashes
					$this_embed = str_replace("\\/", "-escaped-slash-", $this_embed); // temporarily replace escaped slashes so we can separate the search from the replace
					$this_embed = str_replace("-escaped-backslash-", "\\\\", $this_embed); // restore the escaped backslashes
					while (preg_match("/\s+replace\s+\/([^\/]*)\/([^\/]*)\//", $this_embed, $matches)) {
						$search = $matches[1];
						$replace = $matches[2];
						$search = str_replace("-escaped-slash-", "\\/", $search); // restore the escaped slashes
						$replace = str_replace("-escaped-slash-", "\\/", $replace); // restore the escaped slashes
						array_push($this_replace_array, $search);
						array_push($this_replace_array, $replace);
						$this_embed = preg_replace("/\s+replace\s+\/[^\/]*\/[^\/]*\//", "", $this_embed, 1);
					}
					$this_embed_replace = $this_replace_array;
					//$output .= "this_replace_array: ".$embeds_replace[$i - 1]; // debug
				}
				
				$output = preg_replace("/<!--\#\s*embed\s*".preg_quote($this_embed_copy, "/")."\s*-->/i", "-embed-$i-", $output, -1); // this is global so that if the same embed tag appears more than once, we only have to parse it once
				//$output .= "<p>Found embed tag: $this_embed</p>\n\n";  // debug
		
		
// process each embed tag and replace the placeholders with real content

			// show a warning if we have one of our special variable names; you can't name an embed tag with one of these names
		
				if (($this_embed == "page")||($this_embed == "template")||(in_array($this_embed, $reserved_argument_names))) {
					$content = show_error("The tag name \"$this_embed\" is a reserved word<br />".CHANGE_TAG_NAME);
					$this_type = "ignore";
				}
				
		
			// get the args if we've hard-coded them into the tag
		
				if (preg_match("/args[ ]*=[ ]*([^ ]*)/is", $this_embed, $matches)) {
					$this_args = $matches[1];
					while (preg_match("/(\w+)\[(\d*)\]/", $this_args, $matches)) {
						// process passthru/hard-coded combos
						$variable_name = $matches[1];
						$variable_index = $matches[2];
						// these variables don't exist internally as arrays; this fakes it so they're still available in passthru tags
						if (($variable_name == "page")||($variable_name == "template")||($variable_name == "page_definition")) {
							if ((SPLIT_PAGE_NAMES_ON)&&($variable_name == "page")) {
								if ($variable_index == "") {
									// if no index is specified, return the whole page name
									$value = $page;
								} else {
									$page_array = explode(SPLIT_PAGE_NAMES_ON, $page);
									$value = $page_array[$variable_index];
								}
							} else {
								$value = ${$variable_name};
							}
						} else {
							if ($variable_index == "") {
								// if no index is specified, return the whole array as a string
								$value = implode(",", ${$variable_name});
							} else {
								$value = $args[$variable_name][$variable_index];
							}
						}
						$key = $variable_name."[".$variable_index."]";
						$this_args = str_replace($key, $value, $this_args);
					}
					$this_args = preg_replace("/\+/", "", $this_args, -1);
					$this_embed = "hardcoded"; // it doesn't matter what we use for the name here, because...
					$args[$this_embed] = explode(",", $this_args); // ...this is a fake argument that we'll unset as soon as we're done processing it
				}
				
		
			// set the type to passthru if we have a square bracket (and remove the optional $ to avoid errors)
		
				if (preg_match("/(\w+)\[(\d*)\]/", $this_embed, $matches)) {
					$this_embed = $matches[1];
					$this_embed_passthru_index = $matches[2];
					$this_type = "passthru";
				}

		
			// set the type to the name if we have a navigation or environment tag
		
				if ((stristr($this_embed, "nav_"))||(stristr($this_embed, "env_"))) {
					$this_type = $this_embed;
				}
				
		
			// if we don't have any of the previously listed special types, the type is the first argument

				if (!$this_type) {
					if ((isset($args[$this_embed]))&&(is_array($args[$this_embed]))) {
						$this_type = $args[$this_embed][0];
					} else {
						$this_type = "skip";
					}
				}
		
		
			// open the content file, which we can assume to be the second argument
			
				if (($this_type == "file")||(strstr($this_type, "field"))||(strstr($this_type, "random"))||(strstr($this_type, "form"))) {
					$content_file = $args[$this_embed][1];
					$content_file = relative_path($content_file, "content");
					if (!stristr($content_file, "not authorized")) {
						$content = get_file($content_file, "auto", 1);
					} else {
						$content = $content_file;
						$this_type = "error";
					}
					if ($file_missing) { $this_type = "error"; }
				}
		
		
			// get the desired piece of content...
			
			// ...for file
		
				if ($this_type == "file") {
					// (no parsing necessary here)
				}
				
				
			// ...for field
				
				else if ((strstr($this_type, "field"))||(strstr($this_type, "rand"))) {
					$file_type = $args[$this_embed][1]; // determines whether we parse as HTML or XML
					$content_field = (isset($args[$this_embed][2])) ? $args[$this_embed][2] : "" ;
					$content = get_field($content, $content_field, $file_type);
			   }
			   
			   
			// ...for database
				
				else if (strstr($this_type, "database")) {
					$table = $args[$this_embed][1];
					$column = $args[$this_embed][2];
					$key_column = (isset($args[$this_embed][3])) ? $args[$this_embed][3] : "" ;
					$key_value = (isset($args[$this_embed][4])) ? $args[$this_embed][4] : "" ;
					$content = get_database_field($table, $column, $key_column, $key_value);
			   }
			   
			   
			// ...for form
		
				else if (strstr($this_type, "form")) {
					$file_type = $args[$this_embed][1]; // determines whether we parse as HTML or XML
					$content_field = $args[$this_embed][2];
					$form_element = $args[$this_embed][3];
					$content = get_field($content, $content_field, $file_type);       		
					$content = get_form($content, $form_element);
				}
		
		
			// ...for search
				
				else if (strstr($this_type, "search")) {
					$search_string = get_query_argument("search_string", "text");
					$search_files = get_query_argument("search_files", "filename");
					
					// skip if no search string entered
					if ($search_string != "") {
					
						// if we came here from the expanded URL and not the page shortcut, get the page definitions and name now
						if (!$page) {
							$page_definitions = get_page_definitions();
						}
					
						// loop through each content file
						$results = Array();
						$content_files = get_recursive_listing("../content", READ_FILES_RECURSIVELY, $search_files);
						$content_files_count = count($content_files);
						for ($j=0; $j<$content_files_count; $j++) {
							$file_name = $content_files[$j];
							$file_text = get_file(relative_path($file_name, "content"), "filesystem", 0);
							//$output .= "<p>Searching file ".relative_path($file_name, "content")."</p>\n\n"; // debug
			
							// if the file contains the search string...
							if (preg_match("/(.*?)(".preg_quote($search_string, "/").")(.*)/is", $file_text, $matches)) {
			
								// ...create a preview of the found text...
								$results_preview = make_preview($matches[1], $matches[2], $matches[3]);
								
								// ...and count how many times it appears in the field
								$frequency_file = 0;
							
								// then get the field names in the file, and loop through each field
								$content_fields = get_field_names($file_text, $file_name);
								$content_fields_count = count($content_fields);
								for ($k=0; $k<$content_fields_count; $k++) {
									$field_name = $content_fields[$k];
									$field_text = get_field($file_text, $field_name, $file_name);
									// we don't need to reinstate $field_name in PHP, because functions have their own variable scope
								
									// if the field contains the search string...
									$field_text = strip_tags($field_text); // exclude text contained in a tag
									if (preg_match("/(.*?)(".preg_quote($search_string, "/").")(.*)/is", $field_text, $matches)) {
										
										// ...create a preview of the found text...
										$results_preview = make_preview($matches[1], $matches[2], $matches[3]);
										
										// ...and count how many times it appears in the field
										$frequency_field = 0;
										while (stristr($field_text, $search_string)) {
											$frequency_field++;
											$frequency_file++;
											$field_text = preg_replace("/".preg_quote($search_string, "/")."/i", "", $field_text, 1); // we can't use str_replace here, because we only want to match once
										}
			
										// look up the relevant page names, then add the preview and frequency to our results array
										$found_args = "field,$file_name,$field_name";
										$pages = get_page_names_containing($found_args, $page_definitions);
										$pages_count = count($pages);
										for ($l=0; $l<$pages_count; $l++) {
											$results["$pages[$l] : $results_preview"] = $frequency_field;
										}
										
									}
								
								}
								
								// look up the relevant page names, then add the preview and frequency to our results array
								$found_args = "file,$file_name";
								$pages = get_page_names_containing($found_args, $page_definitions);
								$pages_count = count($pages);
								for ($l=0; $l<$pages_count; $l++) {
									$results["$pages[$l] : $results_preview"] = $frequency_file;
								}
								
							}
							
						}
						
						// rank the search results by frequency
						arsort($results); // unfortunately, this sorts by frequency in descending order, then not by name at all; so items of the same frequency will appear in a random order
						$results_keys = array_keys($results);
						$results_values = array_values($results);
						
						//$content = "<table class=\"contemplate_search_results_table\" cellspacing=\"8\" cellpadding=\"0\" border=\"0\">\n";  // these lines provide a different display option
						$content = "";
						$keys_count = count($results_keys);
						for ($j=0; $j < $keys_count; $j++) {
							$this_page = substr($results_keys[$j], 0, strpos($results_keys[$j], " : "));
							$this_preview = substr($results_keys[$j], strpos($results_keys[$j], " : ") + 3);
							$this_frequency = $results_values[$j];
							//$content .= "<tr valign=\"top\">\n<td>$this_frequency</td>\n";
							//$content .= "<td><a href=\"../contemplate/assembler.php?page=$this_page\">contemplate/assembler.php?page=$this_page</a><br />\n$this_preview\n</td>\n</tr>";
							if ($URL_rewriting_flag) {
								$content .= "<p class=\"contemplate_search_results_preview\"><a href=\"../$URL_rewriting_flag/$this_page.html\" class=\"contemplate_search_results_link\">$URL_rewriting_flag/$this_page.html</a><br />\n$this_preview ($this_frequency)</p>\n\n";
							} else {
								$content .= "<p class=\"contemplate_search_results_preview\"><a href=\"../contemplate/assembler.php?page=$this_page\" class=\"contemplate_search_results_link\">contemplate/assembler.php?page=$this_page</a><br />\n$this_preview ($this_frequency)</p>\n\n";
							}
						}
						//$content .= "</table>\n";
						
						if ($content) {
							$content = "<p class=\"contemplate_search_results_heading\">".SEARCH_LABEL_SUCCESS." \"$search_string\":</p>".$content;
						} else {
							$content = "<p class=\"contemplate_search_results_heading\">".SEARCH_LABEL_FAILURE." \"$search_string\"</p>";
						}
					} else {
						$content = "<p class=\"contemplate_search_results_heading\">".SEARCH_LABEL_NONE."</p>";
					}
				}
		
		
			// ...for navigation
			
				else if ((strstr($this_type, "nav_current"))||(strstr($this_type, "nav_total"))||(strstr($this_type, "nav_next"))||(strstr($this_type, "nav_prev"))||(strstr($this_type, "nav_first"))||(strstr($this_type, "nav_last"))) {
					$content = process_tag_nav($this_type);						
				}
		
		
			// ...for environment variable
			
				else if (strstr($this_type, "env_")) {
					$content = process_tag_env($this_type);
										
					// uncomment this to see all available variables
					//while(list($key,$val)=each($_SERVER)) { // debug
					//	$output .= "$key = $val<br />";
					//}
				}
		
		
			// ...for passthrough/passthru
			
				else if (strstr($this_type, "passthr")) {
					// these variables don't exist internally as arrays; this fakes it so they're still available in passthru tags
					if (($this_embed == "page")||($this_embed == "template")||($this_embed == "page_definition")) {
						if ((SPLIT_PAGE_NAMES_ON)&&($this_embed == "page")) {
							if ($this_embed_passthru_index == "") {
								// if no index is specified, return the whole page name
								$content = $page;
							} else {
								$page_array = explode(SPLIT_PAGE_NAMES_ON, $page);
								if (count($page_array) > $this_embed_passthru_index) {
									$content = $page_array[$this_embed_passthru_index];
								} else {
									$content = "";
								}
							}
						} else {
							$content = ${$this_embed};
						}
					} else {
						if (isset($args[$this_embed])) {
							if ($this_embed_passthru_index == "") {
								// if no index is specified, return the whole array as a string
								$content = implode(",", $args[$this_embed]);
							} else {
								if (count($args[$this_embed]) > $this_embed_passthru_index) {
									$content = $args[$this_embed][$this_embed_passthru_index];
								} else {
									$content = "";
								}
							}
						} else {
							$content = "";
						}
					}
				}
		
		
			// ...or for nothing!
				
				else if (strstr($this_type, "skip")) {
					$content = "<!-- Embed tag \"$this_embed\" not parsed because of blank or too few arguments in URL -->";
				}
		
				else if (strstr($this_type, "ignore")) {
					// do nothing at all
				}
		
				else if (!strstr($this_type, "error")) {
					$content = show_error("Can't understand the data type \"$this_type\"<br />Available types: file, field, database, random, form, search, nav_current, nav_total, nav_next, nav_previous, nav_first, nav_last, env_*, or passthru<br />".CHECK_EMBED);
				}
				
				
			// skip the rest of the parsing if this tag generated an error
			
				if (!strstr($content, "Assembler error")) {
				
				
				// remove all markup from the embed content if the strip attribute was included in the embed tag
				
					if ($this_embed_strip) {
						$content = strip_tags($content);
					}
				   
				   
				// do some search-and-replaces if the replace attribute was included in the embed tag
				
					if ($this_embed_replace) {
						$replace_count = count($this_embed_replace) / 2;
						for ($j=0; $j < $replace_count; $j++) {
							$search = $this_embed_replace[$j*2];
							$replace = $this_embed_replace[($j*2)+1];
							$content = preg_replace("/$search/", "$replace", $content, -1);
						}
					}
				   
				   
				}
				
				
			// look for split tags in the content and create variables for those
			// splits are not nestable; if we needed to offer that, it would be easier to process if the end tag also contained the tag name
			// but the current setup is easier to work with because users can rename a split by only changing the opening tag
			
				while (preg_match("/<!--\#\s*split\s+(\w+)\s*-->(.*?)<!--\#\s*end\s+split\s*-->/is", $content, $matches)) {
					$split_tag = $matches[0];
					$split_name = $matches[1];
					$split_contents = $matches[2];
					
					$splits["split_".$split_name] = $split_contents;
					$content = str_replace($split_tag, "\$"."split_".$split_name, $content);
				}
			

			// if requested, add shortcut links to the web editor pages
			
				if ($show_editor_links) {
					if ((strstr($this_type, "field"))||(strstr($this_type, "random"))||(strstr($this_type, "form"))) {
						$editor_links = "";
						$editor_links .= "<p class=\"meditate_links\">";
						$editor_links .= "<a href=\"$editor_index?page=content_edit&path=".str_replace("../", "", $content_file)."&field=$content_field\">Edit</a>";
						$editor_links .= "&nbsp;&bull;&nbsp;";
						$editor_links .= "<a href=\"$editor_index?page=content_history&filter_content=".str_replace("../", "", $content_file).",$content_field\">History</a>";
						$editor_links .= "</p>";
						$content .= $editor_links;
					}
				}
			

			// add the new content to the output...
			
				if (!$this_embed_variable) {
					
					$output = str_replace("-embed-$i-", $content, $output);
					
				} else {
				
				
			// ...or just save it as a variable and delete the embed placeholder and the fake argument we created to process the tag
				
					$embed_variable_keys[count($embed_variable_keys)] = $this_embed;
					$embed_variable_values[count($embed_variable_values)] = $content;
					
					$output = str_replace("-embed-$i-", "", $output);
					unset($args[$this_embed]);
		
				}
				
		
			// finally, start over with the next embed
			
				$last_content = $content;
				$content = "";
				$i++;
			
			}
	
		}


// merge argument values back together, since other scripts won't expect them to be arrays
	
		foreach ($args as $key=>$value) {
			if ($key != "template") { // we left this as a string when we originally parsed the args
				$args[$key] = implode(",", $value);
			}
		}

	
// put content into any embed tag names marked with the variable attribute (we do this after merging argument variables in case they have the same names)
// first we'll rename the "hardcoded" variables if we have more than one

		if (count($embed_variable_keys)) {
			$hardcoded_count = 0;
			$hardcoded_index = 1;
			for ($i=0; $i<count($embed_variable_keys); $i++) {
				if ($embed_variable_keys[$i] == "hardcoded") {
					$hardcoded_count++;
				}
			}
			for ($i=0; $i<count($embed_variable_keys); $i++) {
				if ($hardcoded_count > 1) {
					$embed_variable_keys[$i] = "hardcoded_".$hardcoded_index;
					$hardcoded_index++;
				}
				$args[$embed_variable_keys[$i]] = $embed_variable_values[$i];
			}
		}
	
	
// if we're showing editor links, propagate that request to all the relative links on the page
// also include a return link at the top of the page
	
		if ($show_editor_links) {
			// update URLs like assembled/page.html?foo=bar
			$output = preg_replace("/(['\"][^:\?\&\s'\">]+\.html?\?[^\s'\">]+)/i", "$1&show_editor_links=1&editor_index=$editor_index", $output, -1);
			// update URLs like assembled/page.html
			$output = preg_replace("/(['\"][^:\?\&\s'\">]+\.html?)([^\?])/i", "$1?show_editor_links=1&editor_index=$editor_index$2", $output, -1);
			// update URLs like contemplate/assembler.php?page=page or contemplate/assembler.php?template=default&main=field,main,home
			$output = preg_replace("/(contemplate\/assembler\.php\?[^\s'\">]+)/i", "$1&show_editor_links=1&editor_index=$editor_index", $output, -1);
		
			// add a return link right after the body tag
			$return_link = "";
			$return_link .= "<div class=\"meditate_toolbar\" style=\"background-color: #666666; border-bottom: solid #333333 1px; padding: 8px;\">";
			$return_link .= "<a class=\"meditate_return_link\" href=\"$editor_index?page=content_edit\" style=\"font-family: Arial; font-size: 13px; font-weight: bold; color: #FFFFFF; margin: 0px;\">Return to editor</a>";
			$return_link .= "</div>";
			$output = preg_replace("/(<body[^>]*>\s*)/i", "$1\n$return_link", $output, 1);
		}


	} // end of the "if $page_definition" check


// return the parsed text as an HTML file

	$header_array = explode("\n", $header);
	for ($i=0; $i<count($header_array); $i++) {
		header($header_array[$i]);
	}	
	
	// change to the preferred line ending set in constants.php
	if (LINE_ENDING_FOR_ASSEMBLED_FILES != "\n") {
		$output = str_replace("\n", LINE_ENDING_FOR_ASSEMBLED_FILES, $output);
	}

	// evaluate the output as PHP, unless something tells us to skip it
	if ((!$skip_rendering)&&(isset($args))) {
		$preserved_variables = compact("request", "URL_rewriting_flag", "page", "page_definition", "visitor_id", "error_log", "skip_rendering"); // any variables used after the extract could potentially be overwritten by values in the embedded code; this lets us restore them
		extract($args); // this turns each argument into a local variable so embedded scripts can access them
		extract($splits); // this turns each split into a local variable so embedded scripts can access them
		ob_start(); // this suppresses browser output
		eval("?".">".$output); // this returns the output while rendering the code it contains
		$output = ob_get_contents();
		ob_end_clean();
		extract($preserved_variables); // this restores our own variables to their previous values
	}
	
	// set external link targets if requested in prefs file
	if (EXTERNAL_LINK_TARGET) {
		// remove any quotes from the external link target, and we'll make it match the quoting style of the href value
		$external_link_target_unquoted = preg_replace("/[\"']/", "", EXTERNAL_LINK_TARGET);
		// temporarily remove server-side script tags that might confuse this parsing
		$output = preg_replace("/<\?(.*?)\?>/", "script_tag_start_!@#$1script_tag_end_!@#", $output, -1);
		// add our default target to every anchor tag with an absolute URL (not including absolute URLs to the current domain)
		$output = preg_replace("/<a href=(\\\?[\'\"]?)(\w+:\/\/(?!".preg_quote($_SERVER["HTTP_HOST"]).")[^>]*)>/", "<a href=$1$2 target=$1$external_link_target_unquoted$1>", $output, -1);
		// then remove it from every tag that already had one
		$output = preg_replace("/<a href=(\\\?[\'\"]?)(\w+:\/\/[^>]*target=[^>]*) target=[^>]*>/", "<a href=$1$2>", $output, -1);
		// restore server-side script tags
		$output = preg_replace("/script_tag_start_!@#(.*?)script_tag_end_!@#/", "<"."?$1?".">", $output, -1);
	}
	
	// display the output
	print $output;

	// if requested, display the output in a numbered, pre-rendered format so we can see where an error occurred
	// but only allow this from IP addresses listed in the constants file
	if ((SHOW_SOURCE)&&(isset($_SERVER["REMOTE_ADDR"]))&&(in_string_array($_SERVER["REMOTE_ADDR"], SHOW_SOURCE_TO_IP_ADDRESS))) {
		$output_numbered = $output;
		$output_numbered = str_replace("<", "&lt;", $output_numbered);
		$output_numbered = str_replace(">", "&gt;", $output_numbered);
		$output_numbered = explode(LINE_ENDING_FOR_ASSEMBLED_FILES, $output_numbered);
		for ($i=0; $i<count($output_numbered); $i++) {
			$output_numbered[$i] = "<tr valign=\"top\"><td>".($i+1)."</td><td>&nbsp;</td><td>".$output_numbered[$i]."</td></tr>";
		}
		$output_numbered = implode(LINE_ENDING_FOR_ASSEMBLED_FILES, $output_numbered);
		print "<html>\n<body bgcolor=\"#FFFFFF\">\n<table class=\"contemplate_show_source_table\">\n$output_numbered\n</table>\n</body>\n</html>\n";
	}


// log this page view in the traffic reports

	if ((LOG_SITE_TRAFFIC)&&(!$skip_rendering)) {
	
		if (get_database("reporter")) {
	
			$http_status = (isset($_SERVER["REDIRECT_STATUS"])) ? $_SERVER["REDIRECT_STATUS"] : "200" ;
		
			if ((strstr($request, "contemplate/assembler"))||($URL_rewriting_flag)||($http_status == "200")) {
				// if we recognize the request as a call to the assembler, display either the page name or the expanded URL
				if (!$page) { $page = $page_definition; }
			} else {
				// otherwise, we must have been redirected to here, so show the original request with the error code
				$page = substr($request, 1); // the substr removes the initial slash
			}
			
			// catch the other environment variables now
			$ip_address = (isset($_SERVER["REMOTE_ADDR"])) ? $_SERVER["REMOTE_ADDR"] : "" ;
			$host = (isset($_SERVER["REMOTE_HOST"])) ? $_SERVER["REMOTE_HOST"] : $ip_address ; // if we knew that all servers substituted the IP address for us when the host is missing, then we could remove the conditional here
			$user_agent = (isset($_SERVER["HTTP_USER_AGENT"])) ? $_SERVER["HTTP_USER_AGENT"] : "" ;
			$referrer = (isset($_SERVER["HTTP_REFERER"])) ? $_SERVER["HTTP_REFERER"] : "" ;
			
			// find out whether we need to do our own host lookup
			if ( (HOST_LOOKUPS_FOR_LOGS == "on") || ((HOST_LOOKUPS_FOR_LOGS == "auto")&&($host == $ip_address)) ) {
				$host = gethostbyaddr($ip_address);
			}
	
			// escape any quotes that might be in the error message, for the database query
			$assembler_error = ($error_log) ? addslashes($error_log) : "" ;
			
			// add the visitor ID to each entry if we're using cookies to track visitors
			if (USE_COOKIES_TO_TRACK_VISITORS) {
				if (!isset($visitor_id)||(!$visitor_id)) {
					// get the visitor ID from the cookie if we didn't just generate one on this page
					// if we ever use the cookie to store other data, we can move this code to the top the script so that it's available to the rest of the script; right now it's here so that we can only run it when we need it
					$cookie_data = Array();
					$cookie_data_raw = $_COOKIE["Contemplate"];
					$cookie_data_raw = explode("&", $cookie_data_raw);
					for ($i=0; $i<count($cookie_data_raw); $i++) {
						$cookie_data_item = explode("=", $cookie_data_raw[$i]);
						$cookie_data[$cookie_data_item[0]] = $cookie_data_item[1];
					}
					$visitor_id = $cookie_data["visitor_ID"];
				}
			} else {
				$visitor_id = "";
			}
			
			// if there's a previous page in this session, we'll remove its exit flag and set its duration
			// if this fails, then we have a new session and we'll set the entry value for this page view
			if ((USE_COOKIES_TO_TRACK_VISITORS)&&($visitor_id)) {
				$query = "
					UPDATE `".REPORTER_DATABASE_TABLE."`
					SET `exit`=0, duration=ABS(UNIX_TIMESTAMP()-UNIX_TIMESTAMP(timestamp))
					WHERE visitor_id='$visitor_id'
					AND (UNIX_TIMESTAMP()-UNIX_TIMESTAMP(timestamp))<(".(MINUTES_UNTIL_VISIT_EXPIRES * 60).")
					AND `exit`=1
					ORDER BY timestamp DESC
					LIMIT 1
				";
				$result = mysql_query($query);
				$entry = (mysql_affected_rows()) ? 0 : 1 ;
			} else {
				$entry = 0;
			}
			
			// log this page view
			$query = "INSERT INTO `".REPORTER_DATABASE_TABLE."` (
				timestamp,
				visitor_id,
				ip_address,
				host,
				page,
				user_agent,
				referrer,
				entry,
				`exit`,
				duration,
				http_status,
				assembler_error
			) VALUES (
				 NOW(),
				'$visitor_id',
				'$ip_address',
				'$host',
				'$page',
				'$user_agent',
				'$referrer',
				 $entry,
				 1,
				 0,
				'$http_status',
				'$assembler_error'
			)";
			$result = mysql_query($query);
			
			if (!$result) {
				print "<!-- Can't log the page view to the reporter database -->\n";
			}
		} else {
			print "<!-- Can't connect to the reporter database to log site traffic -->\n";
		}

	}

?>